#include <iomanip>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>     // For (s)random

#include "imms.h"
#include "strmanip.h"

Imms::Imms() : threshold(INITIAL_RATING), last_skipped(true),
    last_jumped(false), dont_rate(true)
{
    srandom(time(NULL));

    fout.open("/dev/null", ofstream::out | ofstream::app);
}

void Imms::setup(const char* _email, const char* _logfile)
{
    fout.close();
    fout.open(_logfile, ofstream::out | ofstream::app);
    email = _email;
    time_t t = time(NULL);
    fout << endl << endl << ctime(&t) << "User: '" << email << "'";
}

bool Imms::song_changed(const string &path, int time_left, bool jumped)
{
    if (access(path.c_str(), R_OK))
    {
        dont_rate = true;
        return false;
    }

    // Rate the old song
    if (!dont_rate)
    {
        int mod;

        if (time_left > 2)
        {
            if (last_jumped && !jumped)
                mod = JUMPED_TO;
            else if (jumped)
                mod = JUMPED_FROM;
            else if (last_skipped)
                mod = CONS_SKIP_RATE;
            else
                mod = FIRST_SKIP_RATE;
            last_skipped = true;
        }
        else
        {
            if (last_jumped)
                mod = JUMPED_TO_FIN;
            else if (last_skipped)
                mod = FIRST_NON_SKIP_RATE;
            else
                mod = CONS_NON_SKIP_RATE;
            last_skipped = jumped = false;
        }

        fout << (last_jumped ? "[Jump: To] " : "");
        fout << (jumped ? "[Jump: From] " : "");
        fout << (last_skipped && !jumped ? "[Skip: Manual] " : "");
        fout << "[Delta: " << setiosflags(ios::showpos) << mod <<
            resetiosflags (ios::showpos) << "] ";

        last_jumped = jumped;
    
        set_popm(mod);
    }
    else
    {
        dont_rate = last_jumped = jumped = false;
    }

    // Start processing the new song

    fout << endl << string(TERM_WIDTH - 20, '-') << endl;

    if (path.length() > TERM_WIDTH - 8)
        fout << "[Path: ..." << path.substr(path.length() - TERM_WIDTH + 11,
                TERM_WIDTH - 11) << "]" << endl;
    else
        fout << "[Path: " << path << "]" << endl;

    id3tag.Clear();

    // Don't write tags to funny file types - the decoder might get confused
    if (path.length() < 3)
    {
        fout << "[NotMP3] ";
        dont_rate = true; 
        return false;
    }

    string ext = path.substr(path.length() - 3, 3);
    string_tolower(ext);

    if (ext != "mp3" && ext != "ogg")
    {
        fout << "[NotMP3] ";
        dont_rate = true;
        return false;
    }
    else if (access(path.c_str(), W_OK))
    {
        fout << "[RdOnly] ";
        dont_rate = true;
        return false;
    }

    id3tag.Link(path.c_str());

    fout << "[Cutoff: " << setw(5) << threshold << "] ";
    fout << "[Rating: " << setw(3) << get_popm() << "] ";

    if (!jumped && threshold > get_popm())
    {
        // Figure out how far below the threshold this song falls
        float total = pow(2, (threshold - get_popm())/EXPONENT_STEP) + BIAS;
        fout << "[Chance: " << setprecision(2) << 1 / total << "] ";
        // remember to reset precision
        fout << setprecision(4);
        if (random() % 100 * total / 99 > 1) 
        {
            fout << "[Skip: Auto] ";
            dont_rate = true;
        }
    }
    if (get_popm() > threshold)
        threshold = threshold * (1 - THRESHOLD_GROW) +
            get_popm() * THRESHOLD_GROW;
    else 
        threshold = threshold * (1 - THRESHOLD_DECAY) +
            get_popm() * THRESHOLD_DECAY;

    return dont_rate;
}

void Imms::set_popm(int mod)
{
    ID3_Frame* popmframe = id3tag.Find(ID3FID_POPULARIMETER,
            ID3FN_EMAIL, email.c_str());

    if (!popmframe)
    {
        fout << "[New] "; 
        popmframe = new ID3_Frame(ID3FID_POPULARIMETER);
        popmframe->GetField(ID3FN_EMAIL)->Set(email.c_str()); 
        id3tag.AttachFrame(popmframe); // id3 takes ownership
        popmframe = id3tag.Find(ID3FID_POPULARIMETER,
                ID3FN_EMAIL, email.c_str());
    }

    ID3_Field* ratingfield = popmframe->GetField(ID3FN_RATING);

    int newrating = get_popm() + mod;

    if (newrating > 150)
        newrating = 150;

    if (newrating < 50)
        newrating = 50;

    ratingfield->Set(newrating);

    id3tag.Update();
}

int Imms::get_popm()
{
    ID3_Frame* popmframe = id3tag.Find(ID3FID_POPULARIMETER,
            ID3FN_EMAIL, email.c_str());

    if (popmframe)
    {
        ID3_Field* ratingfield = popmframe->GetField(ID3FN_RATING);
        if (ratingfield && ratingfield->Get() > 0)
        {
            return ratingfield->Get();
        }
    }
    return INITIAL_RATING;
}
